/* aztec.c - Handles Aztec Mesa 2D Symbols */

/*
    libzint - the open source barcode library
    Copyright (C) 2008 Robin Stuart <zint@hotmail.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include "common.h"
#include "aztec.h"
#include "reedsol.h"
//#include "zint.h"
#define SAFE_FREE(a) if(a){free(a);a=NULL;}

/*
	This code encodes according to the system specified in US Patent 5,591,956
	entitled "Two Dimensional Data Encoding Structure and Symbology For Use With
	Optical Readers" by Andrew Longacre, Jr. and Rob Hussey. The patent
	includes the following dedication:
	
	"5,591,956 - Andrew Longacre, Jr., Skaneateles; Rob Husset, Liverpool, both
	of N.Y. TWO DIMENIONAL DATA ENCODING STRUCTURE AND SYMBOLOGY FOR USE WITH
	OPTICAL READERS. Patent dated Jan. 7, 1997. Dedication filed March 20, 1997,
	by the assignee, Welch Allyn, Inc. Hereby dedicates to the public the entire
	term of said patent. (Official Gazette, June 17, 1997)"
*/

void mapshorten(int *charmap, int *typemap, int start, int length)
{ /* Shorten the string by one character */
	int i;
	
	for(i = start + 1; i < (length - 1); i++) {
		charmap[i] = charmap[i + 1];
		typemap[i] = typemap[i + 1];
	}
}

void insert(char binary_string[], int posn, char newbit)
{ /* Insert a character into the middle of a string at position posn */
	int i, end;
	
	end = strlen(binary_string);
	for(i = end; i > posn; i--) {
		binary_string[i] = binary_string[i - 1];
	}
	binary_string[posn] = newbit;
}

int aztec_text_process(unsigned char source[], char binary_string[])
{

	/* Encode input data into a binary string */
	int map_posn, i, j, k, bytes;
	//** 
	//int charmap[strlen(source)], typemap[strlen(source)], maplength;
	//int blockmap[2][strlen(source)], blocks;
	int  maplength, blocks;
	int *charmap=NULL, *typemap=NULL, *blockmap1=NULL, *blockmap2=NULL;
	//**
	int curtable, newtable, lasttable;

	charmap = (int *)malloc((strlen(source)*sizeof(int)));
	typemap = (int *)malloc((strlen(source)*sizeof(int)));
	blockmap1 = (int *)malloc((strlen(source)*sizeof(int)));
	blockmap2 = (int *)malloc((strlen(source)*sizeof(int)));

	/* Lookup input string in encoding table */
	for(i = 0; i < strlen(source); i++) {
		if(source[i] > 127) {
			charmap[i] = source[i];
			typemap[i] = BINARY;
		} else {
			charmap[i] = AztecSymbolChar[source[i]];
			typemap[i] = AztecCodeSet[source[i]];
		}
	}
	maplength = strlen(source);
	
	/* Look for double character encoding possibilities */
	i = 0;
	do{
		if(((charmap[i] == 28) && (charmap[i + 1] == 26)) && ((typemap[i] == PUNC) && (typemap[i + 1] == PUNC))) {
			/* CR LF combination */
			charmap[i] = 0;
			typemap[i] = 12;
			mapshorten(charmap, typemap, i, maplength);
			maplength--;
		}
		
		if(((charmap[i] == 302) && (charmap[i + 1] == 300)) && ((typemap[i] == 28) && (typemap[i + 1] == 23))) {
			/* . SP combination */
			charmap[i] = 1;
			typemap[i] = 12;
			mapshorten(charmap, typemap, i, maplength);
			maplength--;
		}
		
		if(((charmap[i] == 301) && (charmap[i + 1] == 300)) && ((typemap[i] == 28) && (typemap[i + 1] == 23))) {
			/* , SP combination */
			charmap[i] = 2;
			typemap[i] = 12;
			mapshorten(charmap, typemap, i, maplength);
			maplength--;
		}
		
		if(((charmap[i] == 303) && (charmap[i + 1] == 300)) && ((typemap[i] == 12) && (typemap[i + 1] == 23))) {
			/* : SP combination */
			charmap[i] = 3;
			typemap[i] = 12;
			mapshorten(charmap, typemap, i, maplength);
			maplength--;
		}
		
		i++;
	}while(i < maplength);
	
	/* look for blocks of characters which use the same table */
	blocks = 0;
	blockmap1[0] = typemap[0];
	blockmap2[0] = 1;
	for(i = 1; i < maplength; i++) {
		if(typemap[i] == typemap[i - 1]) {
			blockmap2[blocks]++;
		} else {
			blocks++;
			blockmap1[blocks] = typemap[i];
			blockmap2[blocks] = 1;
		}
	}
	
	if(blockmap1[0] & 1) { blockmap1[0] = 1; }
	if(blockmap1[0] & 2) { blockmap1[0] = 2; }
	if(blockmap1[0] & 4) { blockmap1[0] = 4; }
	if(blockmap1[0] & 8) { blockmap1[0] = 8; }
	
	/* look for adjacent blocks which can use the same table (left to right search) */
	for(i = 1; i < blocks; i++) {
		if(blockmap1[i] & blockmap1[i - 1]) {
			blockmap1[i] = (blockmap1[i] & blockmap1[i - 1]);
		}
	}
	
	if(blockmap1[blocks - 1] & 1) { blockmap1[blocks - 1] = 1; }
	if(blockmap1[blocks - 1] & 2) { blockmap1[blocks - 1] = 2; }
	if(blockmap1[blocks - 1] & 4) { blockmap1[blocks - 1] = 4; }
	if(blockmap1[blocks - 1] & 8) { blockmap1[blocks - 1] = 8; }
	
	/* look for adjacent blocks which can use the same table (right to left search) */
	for(i = blocks; i > 0; i--) {
		if(blockmap1[i] & blockmap1[i + 1]) {
			blockmap1[i] = (blockmap1[i] & blockmap1[i + 1]);
		}
	}
	
	/* determine the encoding table for characters which do not fit with adjacent blocks */
	for(i = 1; i < blocks; i++) {
		if(blockmap1[i] & 8) { blockmap1[i] = 8; }
		if(blockmap1[i] & 4) { blockmap1[i] = 4; }
		if(blockmap1[i] & 2) { blockmap1[i] = 2; }
		if(blockmap1[i] & 1) { blockmap1[i] = 1; }
	}
	
	/* Combine blocks of the same type */
	blocks;
	i = 0;
	do{
		if(blockmap1[i] == blockmap1[i + 1]) {
			blockmap2[i] += blockmap2[i + 1];
			for(j = i + 1; j < blocks; j++) {
				blockmap1[j] = blockmap1[j + 1];
				blockmap2[j] = blockmap2[j + 1];
			}
			blocks--;
		} else {
			i++;
		}
	} while (i < blocks);
	
	/* Put the adjusted block data back into typemap */
	j = 0;
	for(i = 0; i <= blocks; i++) {
		if((blockmap2[i] < 3) && (blockmap1[i] != 32)) { /* Shift character(s) needed */
			for(k = 0; k < blockmap2[i]; k++) {
				typemap[j + k] = blockmap1[i] + 64;
			}
		} else { /* Latch character (or byte mode) needed */
			for(k = 0; k < blockmap2[i]; k++) {
				typemap[j + k] = blockmap1[i];
			}
		}
		j += blockmap2[i];
	}
	
	/* Don't shift an initial capital letter */
	if(typemap[0] == 65) { typemap[0] = 1; };
	
	/* Problem characters (those that appear in different tables with different values) can now be resolved into their tables */
	for(i = 0; i < maplength; i++) {
		if(charmap[i] >= 300) {
			curtable = typemap[i];
			if(curtable > 64) {
				curtable -= 64;
			}
			switch(charmap[i]) {
				case 300: /* Space */
					switch(curtable) {
						case UPPER: charmap[i] = 26; break;
						case LOWER: charmap[i] = 26; break;
						case STRING: charmap[i] = 26; break;
						case DIGIT: charmap[i] = 10; break;
					}
					break;
					case 301: /* Comma */
						switch(curtable) {
							case STRING: charmap[i] = 15; break;
							case PUNC: charmap[i] = 15; break;
							case DIGIT: charmap[i] = 11; break;
						}
						break;
					case 302: /* Full Stop */
						switch(curtable) {
							case STRING: charmap[i] = 11; break;
							case PUNC: charmap[i] = 17; break;
							case DIGIT: charmap[i] = 12; break;
						}
						break;
					case 303: /* Colon */
						switch(curtable) {
							case STRING: charmap[i] = 12; break;
							case PUNC: charmap[i] = 19; break;
						}
						break;
			}
		}
	}
	
	for(i = 0; i < 20000; i++) {
		binary_string[i] = '\0';
	}
	
	curtable = UPPER; /* start with UPPER table */
	for(i = 0; i < maplength; i++) {
		newtable = curtable;
		if(typemap[i] != curtable) {
			/* Change table */
			if(curtable == BINARY) {
				/* If ending binary mode the current table is the same as when entering binary mode */
				curtable = lasttable;
				newtable = lasttable;
			}
			if(typemap[i] > 64) {
				/* Shift character */
				switch(typemap[i]) {
					case (64 + UPPER): /* To UPPER */
						switch(curtable) {
							case LOWER: /* US */
								concat(binary_string, hexbit[27]);
								break;
							case STRING: /* UL */
								concat(binary_string, hexbit[28]);
								newtable = UPPER;
								break;
							case PUNC: /* UL */
								concat(binary_string, hexbit[28]);
								newtable = UPPER;
								break;
							case DIGIT: /* US */
								concat(binary_string, pentbit[13]);
								break;
						}
						break;
					case (64 + LOWER): /* To LOWER */
						switch(curtable) {
							case UPPER: /* LL */
								concat(binary_string, hexbit[27]);
								newtable = LOWER;
								break;
							case STRING: /* LL */
								concat(binary_string, hexbit[27]);
								newtable = LOWER;
								break;
							case PUNC: /* UL LL */
								concat(binary_string, hexbit[31]);
								concat(binary_string, hexbit[27]);
								newtable = LOWER;
								break;
							case DIGIT: /* UL LL */
								concat(binary_string, pentbit[14]);
								concat(binary_string, hexbit[27]);
								newtable = LOWER;
								break;
						}
						break;
					case (64 + STRING): /* To STRING */
						switch(curtable) {
							case UPPER: /* SL */
								concat(binary_string, hexbit[28]);
								newtable = STRING;
								break;
							case LOWER: /* SL */
								concat(binary_string, hexbit[28]);
								newtable = STRING;
								break;
							case PUNC: /* UL SL */
								concat(binary_string, hexbit[31]);
								concat(binary_string, hexbit[28]);
								newtable = STRING;
								break;
							case DIGIT: /* UL SL */
								concat(binary_string, pentbit[14]);
								concat(binary_string, hexbit[28]);
								newtable = STRING;
								break;
						}
						break;
					case (64 + PUNC): /* To PUNC */
						switch(curtable) {
							case UPPER: /* PS */
								concat(binary_string, hexbit[29]);
								break;
							case LOWER: /* PS */
								concat(binary_string, hexbit[29]);
								break;
							case STRING: /* PS */
								concat(binary_string, hexbit[29]);
								break;
							case DIGIT: /* PS */
								concat(binary_string, pentbit[15]);
								break;
						}
						break;
					case (64 + DIGIT): /* To DIGIT */
						switch(curtable) {
							case UPPER: /* DL */
								concat(binary_string, hexbit[30]);
								newtable = DIGIT;
								break;
							case LOWER: /* DL */
								concat(binary_string, hexbit[30]);
								newtable = DIGIT;
								break;
							case STRING: /* UL DL */
								concat(binary_string, hexbit[28]);
								concat(binary_string, hexbit[30]);
								newtable = DIGIT;
								break;
							case PUNC: /* UL DL */
								concat(binary_string, hexbit[31]);
								concat(binary_string, hexbit[30]);
								newtable = DIGIT;
								break;
						}
						break;
				}
			} else {
				/* Latch character */
				switch(typemap[i]) {
					case UPPER: /* To UPPER */
						switch(curtable) {
							case LOWER: /* SL UL */
								concat(binary_string, hexbit[28]);
								concat(binary_string, hexbit[28]);
								newtable = UPPER;
								break;
							case STRING: /* UL */
								concat(binary_string, hexbit[28]);
								newtable = UPPER;
								break;
							case PUNC: /* UL */
								concat(binary_string, hexbit[28]);
								newtable = UPPER;
								break;
							case DIGIT: /* UL */
								concat(binary_string, pentbit[14]);
								newtable = UPPER;
								break;
						}
						break;
					case LOWER: /* To LOWER */
						switch(curtable) {
							case UPPER: /* LL */
								concat(binary_string, hexbit[27]);
								newtable = LOWER;
								break;
							case STRING: /* LL */
								concat(binary_string, hexbit[27]);
								newtable = LOWER;
								break;
							case PUNC: /* UL LL */
								concat(binary_string, hexbit[31]);
								concat(binary_string, hexbit[27]);
								newtable = LOWER;
								break;
							case DIGIT: /* UL LL */
								concat(binary_string, pentbit[14]);
								concat(binary_string, hexbit[27]);
								newtable = LOWER;
								break;
						}
						break;
					case STRING: /* To STRING */
						switch(curtable) {
							case UPPER: /* SL */
								concat(binary_string, hexbit[28]);
								newtable = STRING;
								break;
							case LOWER: /* SL */
								concat(binary_string, hexbit[28]);
								newtable = STRING;
								break;
							case PUNC: /* UL SL */
								concat(binary_string, hexbit[31]);
								concat(binary_string, hexbit[28]);
								newtable = STRING;
								break;
							case DIGIT: /* UL SL */
								concat(binary_string, pentbit[14]);
								concat(binary_string, hexbit[28]);
								newtable = STRING;
								break;
						}
						break;
					case PUNC: /* To PUNC */
						switch(curtable) {
							case UPPER: /* SL PL */
								concat(binary_string, hexbit[28]);
								concat(binary_string, hexbit[30]);
								newtable = PUNC;
								break;
							case LOWER: /* SL PL */
								concat(binary_string, hexbit[28]);
								concat(binary_string, hexbit[30]);
								newtable = PUNC;
								break;
							case STRING: /* PL */
								concat(binary_string, hexbit[30]);
								newtable = PUNC;
								break;
							case DIGIT: /* UL SL PL */
								concat(binary_string, pentbit[14]);
								concat(binary_string, hexbit[28]);
								concat(binary_string, hexbit[30]);
								newtable = PUNC;
								break;
						}
						break;
					case DIGIT: /* To DIGIT */
						switch(curtable) {
							case UPPER: /* DL */
								concat(binary_string, hexbit[30]);
								newtable = DIGIT;
								break;
							case LOWER: /* DL */
								concat(binary_string, hexbit[30]);
								newtable = DIGIT;
								break;
							case STRING: /* UL DL */
								concat(binary_string, hexbit[28]);
								concat(binary_string, hexbit[30]);
								newtable = DIGIT;
								break;
							case PUNC: /* UL DL */
								concat(binary_string, hexbit[31]);
								concat(binary_string, hexbit[20]);
								newtable = DIGIT;
								break;
						}
						break;
					case BINARY: /* To BINARY */
						lasttable = curtable;
						switch(curtable) {
							case UPPER: /* BS */
								concat(binary_string, hexbit[31]);
								newtable = BINARY;
								break;
							case LOWER: /* BS */
								concat(binary_string, hexbit[31]);
								newtable = BINARY;
								break;
							case STRING: /* BS */
								concat(binary_string, hexbit[31]);
								newtable = BINARY;
								break;
							case PUNC: /* UL BS */
								concat(binary_string, hexbit[31]);
								concat(binary_string, hexbit[31]);
								newtable = BINARY;
								break;
							case DIGIT: /* UL BS */
								concat(binary_string, pentbit[14]);
								concat(binary_string, hexbit[31]);
								newtable = BINARY;
								break;
						}
		
						bytes = 0;
						do{
							bytes++;
						}while(typemap[i + (bytes - 1)] == BINARY);
		
						if(bytes > 2079) {
							//** 
							SAFE_FREE(charmap);
							SAFE_FREE(typemap);
							SAFE_FREE(blockmap1);
							SAFE_FREE(blockmap2);
							//**
							return ERROR_TOO_LONG;
						}
		
						if(bytes > 31) { /* Put 00000 followed by 11-bit number of bytes less 31 */
							int adjusted;
			
							adjusted = bytes - 31;
							concat(binary_string, "00000");
							if(adjusted & 0x400) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x200) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x100) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x80) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x40) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x20) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x10) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x08) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x04) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x02) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(adjusted & 0x01) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
						} else { /* Put 5-bit number of bytes */
							if(bytes & 0x10) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(bytes & 0x08) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(bytes & 0x04) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(bytes & 0x02) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
							if(bytes & 0x01) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
						}
		
						break;
				}
			}
		}
		/* Add data to the binary string */
		curtable = newtable;
		switch(curtable) {
			case UPPER:
			case LOWER:
			case STRING:
			case PUNC:
				concat(binary_string, hexbit[charmap[i]]);
				break;
			case DIGIT:
				concat(binary_string, pentbit[charmap[i]]);
				break;
			case BINARY:
				if(charmap[i] & 0x80) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
				if(charmap[i] & 0x40) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
				if(charmap[i] & 0x20) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
				if(charmap[i] & 0x10) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
				if(charmap[i] & 0x08) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
				if(charmap[i] & 0x04) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
				if(charmap[i] & 0x02) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
				if(charmap[i] & 0x01) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
				break;
		}
		
		if(strlen(binary_string) > 14970) {
			//** 
			SAFE_FREE(charmap);
			SAFE_FREE(typemap);
			SAFE_FREE(blockmap1);
			SAFE_FREE(blockmap2);
			//**
			return ERROR_TOO_LONG;
		}
	}
	
	/* Data string can't have 10 '0's or 10 '1's in a 10-bit block */
	i = 0;
	do{
		int count0, count1;
		
		count0 = 0;
		count1 = 0;
		for(j = 0; j < 10; j++) {
			if(binary_string[j + i] == '0') { count0++; } else { count1++; }
		}
		if(count0 == 10) { /* If all 0s insert a '1' */
			insert(binary_string, i+9, '1');
		}
		if(count1 == 10) { /* If all 1s insert a '0' */
			insert(binary_string, i+9, '0');
		}
		i += 10;
	} while (i < strlen(binary_string));
	
	/* Insert extra '1's to pad out the final 10-bit block of data */
	k = strlen(binary_string);
	for(j = 0; j < (i - k); j++) {
		concat(binary_string, "1");
	}
	
	/* But be careful that the final 10-bit block of data doesn't contain all '1's */
	k = 0;
	for(j = (i - 10); j < i; j++) {
		if(binary_string[j] == '1') { k++; }
	}
	if(k == 10) { /* Change the very last bit */
		binary_string[strlen(binary_string)] = '0';
	}
	
	//** 
	SAFE_FREE(charmap);
	SAFE_FREE(typemap);
	//free(blockmap1);
	//blockmap1=NULL;
	//free(blockmap2);
	//blockmap2=NULL;
	SAFE_FREE(blockmap1);
	SAFE_FREE(blockmap2);
	//**
	return 0;
}


//---

int aztec_mesa(struct zint_symbol *symbol, unsigned char source[])
{
	int x, y, i, data_blocks, ecc_blocks, layers, total_bits;
	char binary_string[20000], bit_pattern[20045], descriptor[42];
	unsigned int data_part[1500], ecc_part[510];
	unsigned char desc_data[4], desc_ecc[6];
	
	if(aztec_text_process(source, binary_string) != 0) {
		strcpy(symbol->errtxt, "Input too long or too many extended ASCII characters");
		return ERROR_TOO_LONG;
	}
	
	data_blocks = strlen(binary_string) / 10;
	
	layers = 32;
	for(i = 31; i >= 0; i--) {
		if(AztecDataSizes[i] > data_blocks) { layers = (i + 1); }
	}
	ecc_blocks = AztecSizes[layers - 1] - data_blocks;
	
	/* Copy across data into separate integers */
	for(i = 0; i < 1500; i++) {
		data_part[i] = 0;
	}
	for(i = 0; i < 510; i++) {
		ecc_part[i] = 0;
	}
	
	for(i = 0; i < data_blocks; i++) {
		if(binary_string[i * 10] == '1') { data_part[i] += 512; }
		if(binary_string[(i * 10) + 1] == '1') { data_part[i] += 256; }
		if(binary_string[(i * 10) + 2] == '1') { data_part[i] += 128; }
		if(binary_string[(i * 10) + 3] == '1') { data_part[i] += 64; }
		if(binary_string[(i * 10) + 4] == '1') { data_part[i] += 32; }
		if(binary_string[(i * 10) + 5] == '1') { data_part[i] += 16; }
		if(binary_string[(i * 10) + 6] == '1') { data_part[i] += 8; }
		if(binary_string[(i * 10) + 7] == '1') { data_part[i] += 4; }
		if(binary_string[(i * 10) + 8] == '1') { data_part[i] += 2; }
		if(binary_string[(i * 10) + 9] == '1') { data_part[i] += 1; }
	}
	
	/* Add reed-solomon codes */
	/* "The check blocks are calculated by systematic Reed-Solomon encoding using the
	Galois Field GF(1024) with a prime modulus polynomial of x**10 + x**3 + 1." (column 9, line 37) */
	rs_init_gf(0x205);
	rs_init_code(ecc_blocks, 1);
	rs_encode_long(data_blocks, data_part, ecc_part);
	
	for(i = 0; i < ecc_blocks; i++) {
		if(ecc_part[i] & 0x200) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
		if(ecc_part[i] & 0x100) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
		if(ecc_part[i] & 0x80) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
		if(ecc_part[i] & 0x40) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
		if(ecc_part[i] & 0x20) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
		if(ecc_part[i] & 0x10) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
		if(ecc_part[i] & 0x08) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
		if(ecc_part[i] & 0x04) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
		if(ecc_part[i] & 0x02) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
		if(ecc_part[i] & 0x01) { concat(binary_string, "1"); } else { concat(binary_string, "0"); }
	}
	
	/* Invert the data so that actual data is on the outside and reed-solomon on the inside */
	for(i = 0; i < 20045; i++) {
		bit_pattern[i] = '0';
	}
	
	total_bits = (data_blocks + ecc_blocks) * 10;
	for(i = 0; i < total_bits; i++) {
		bit_pattern[i] = binary_string[total_bits - i];
	}
	
	/* Now add the symbol descriptor */
	for(i = 0; i < 42; i++) {
		descriptor[i] = '0';
	}
	
	/* The first 5 bits represent the number of layers minus 1 */
	if((layers - 1) & 0x10) { descriptor[0] = '1'; }
	if((layers - 1) & 0x08) { descriptor[1] = '1'; }
	if((layers - 1) & 0x04) { descriptor[2] = '1'; }
	if((layers - 1) & 0x02) { descriptor[3] = '1'; }
	if((layers - 1) & 0x01) { descriptor[4] = '1'; }
	/* The next 11 bits represent the number of data blocks minus 1 */
	if((data_blocks - 1) & 0x400) { descriptor[5] = '1'; }
	if((data_blocks - 1) & 0x200) { descriptor[6] = '1'; }
	if((data_blocks - 1) & 0x100) { descriptor[7] = '1'; }
	if((data_blocks - 1) & 0x80) { descriptor[8] = '1'; }
	if((data_blocks - 1) & 0x40) { descriptor[9] = '1'; }
	if((data_blocks - 1) & 0x20) { descriptor[10] = '1'; }
	if((data_blocks - 1) & 0x10) { descriptor[11] = '1'; }
	if((data_blocks - 1) & 0x08) { descriptor[12] = '1'; }
	if((data_blocks - 1) & 0x04) { descriptor[13] = '1'; }
	if((data_blocks - 1) & 0x02) { descriptor[14] = '1'; }
	if((data_blocks - 1) & 0x01) { descriptor[15] = '1'; }
	/* "The remaining 24 bits of the symbol descriptor structure contain check characters by
	systematic Reed-Solomon encoding using the Galois Field GF(16) with a prime modulus
	polynomial of x**4 + x + 1." (column 7, line 6) */
	for(i = 0; i < 4; i++) {
		desc_data[i] = 0;
		desc_ecc[i] = 0;
	}
	
	for(i = 0; i < 4; i++) {
		if(descriptor[i * 4] == '1') { desc_data[i] += 8; }
		if(descriptor[(i * 4) + 1] == '1') { desc_data[i] += 4; }
		if(descriptor[(i * 4) + 2] == '1') { desc_data[i] += 2; }
		if(descriptor[(i * 4) + 3] == '1') { desc_data[i] += 1; }
	}
	
	rs_init_gf(0x13);
	rs_init_code(6, 1);
	rs_encode(4, desc_data, desc_ecc);
	for(i = 0; i < 6; i++) {
		if(desc_ecc[i] & 0x08) { descriptor[(i * 4) + 16] = '1'; }
		if(desc_ecc[i] & 0x04) { descriptor[(i * 4) + 17] = '1'; }
		if(desc_ecc[i] & 0x02) { descriptor[(i * 4) + 18] = '1'; }
		if(desc_ecc[i] & 0x01) { descriptor[(i * 4) + 19] = '1'; }
	}
	
	/* Merge descriptor with the rest of the symbol */
	for(i = 0; i < 40; i++) {
		bit_pattern[19999 + i] = descriptor[i];
	}
	
	/* Plot all of the data into the symbol in pre-defined spiral pattern */
	for(y = AztecOffset[layers - 1]; y < (152 - AztecOffset[layers - 1]); y++) {
		for(x = AztecOffset[layers - 1]; x < (152 - AztecOffset[layers - 1]); x++) {
			if(AztecMap[(y * 151) + x] == 1) {
				symbol->encoded_data[y - AztecOffset[layers - 1]][x - AztecOffset[layers - 1]] = '1';
			}
			if(AztecMap[(y * 151) + x] >= 2) {
				if(bit_pattern[AztecMap[(y * 151) + x] - 2] == '1') {
					symbol->encoded_data[y - AztecOffset[layers - 1]][x - AztecOffset[layers - 1]] = '1';
				}
			}
		}
		symbol->row_height[y - AztecOffset[layers - 1]] = 1;
	}
	symbol->rows = 151 - (2 * AztecOffset[layers - 1]);
	symbol->width = 151 - (2 * AztecOffset[layers - 1]);
	
	return 0;
}


